@@ -75,17 +75,17 @@ static Function functions[NFUNCTIONS] = {
 { FN_START_BLOCK,       "(", 0 },
 { FN_END_BLOCK,         ")", 0 },
 { FN_ADD,               "+", 0 },
-{ FN_SUBTRACT,          "-", 0 },
-{ FN_MULTIPLY,          "*", 0 },
-{ FN_DIVIDE,            "/", 0 },
+{ FN_SUBTRACT,          "−", 0 },
+{ FN_MULTIPLY,          "×", 0 },
+{ FN_DIVIDE,            "÷", 0 },
 { FN_BACKSPACE,         NULL, 0 },
 { FN_DELETE,            NULL, 0 },
 { FN_CHANGE_SIGN,       NULL, 0 },
 { FN_INTEGER,           "Int", FUNC },
 { FN_FRACTION,          "Frac", FUNC },
 { FN_PERCENTAGE,        "%", 0 },
-{ FN_SQUARE,            "^2", 0 },
-{ FN_SQUARE_ROOT,       "Sqrt", FUNC },
+{ FN_SQUARE,            "²", 0 },
+{ FN_SQUARE_ROOT,       "√", 0 },
 { FN_RECIPROCAL,        NULL, 0 },
 { FN_E_POW_X,           "e^", PREFIXOP },
 { FN_10_POW_X,          "10^", PREFIXOP },       
@@ -94,26 +94,26 @@ static Function functions[NFUNCTIONS] = {
 { FN_X_POW_Y_INV,       "^(1/(", 0 },
 { FN_FACTORIAL,         "!", 0 },
 { FN_RANDOM,            "Rand", 0 },
-{ FN_SIN,               "Sin", FUNC },
-{ FN_SINH,              "Sinh", FUNC },
-{ FN_ASIN,              "Asin", FUNC },
-{ FN_ASINH,             "Asinh", FUNC },
-{ FN_COS,               "Cos", FUNC },
-{ FN_COSH,              "Cosh", FUNC },
-{ FN_ACOS,              "Acos", FUNC },
-{ FN_ACOSH,             "Acosh", FUNC },
-{ FN_TAN,               "Tan", FUNC },
-{ FN_TANH,              "Tanh", FUNC },
-{ FN_ATAN,              "Atan", FUNC },
-{ FN_TAN,               "Atanh", FUNC },
-{ FN_NATURAL_LOGARITHM, "Ln", FUNC },
-{ FN_LOGARITHM,         "Log", FUNC },
-{ FN_LOGARITHM2,        "Log2", FUNC },
+{ FN_SIN,               " sin ", 0 },
+{ FN_SINH,              " sinh ", 0 },
+{ FN_ASIN,              " sin⁻¹ ", 0 },
+{ FN_ASINH,             " asinh ", 0 },
+{ FN_COS,               " cos ", 0 },
+{ FN_COSH,              " cosh ", 0 },
+{ FN_ACOS,              " cos⁻¹ ", 0 },
+{ FN_ACOSH,             " acosh ", 0 },
+{ FN_TAN,               " tan ", 0 },
+{ FN_TANH,              " tanh ", 0 },
+{ FN_ATAN,              " tan⁻¹ ", 0 },
+{ FN_TAN,               " atanh ", 0 },
+{ FN_NATURAL_LOGARITHM, " ln ", 0 },
+{ FN_LOGARITHM,         " log ", 0 },
+{ FN_LOGARITHM2,        " log₂ ", 0 },
 { FN_ABSOLUTE_VALUE,    "Abs", FUNC },
 { FN_TRUNC,             "Trunc", FUNC },
-{ FN_MODULUS_DIVIDE,    " Mod ", 0 },
-{ FN_1S_COMPLEMENT,     "1s", FUNC },
-{ FN_2S_COMPLEMENT,     "2s", FUNC },
+{ FN_MODULUS_DIVIDE,    " mod ", 0 },
+{ FN_1S_COMPLEMENT,     "ones", FUNC },
+{ FN_2S_COMPLEMENT,     "twos", FUNC },
 { FN_EXPONENTIAL,       "e", 0 },
 { FN_NOT,               "~", 0 },
 { FN_OR,                " OR ", 0 },
@@ -177,7 +177,7 @@ do_function(int index)
     assert(index >= 0);
     assert(index <= 9);
 
-    ret = mp_equation_udf_parse(function_get_value(index));
+    ret = 0;//FIXME: mp_equation_udf_parse(function_get_value(index));
     if (!ret) {
         ui_set_statusbar("", "");
     } else {
@@ -189,7 +189,7 @@ do_function(int index)
 
 
 static void
-do_paste(const char *text)
+do_paste(int cursor_start, int cursor_end, const char *text)
 {
     const char *input;
     char c, *output, *clean_text;
@@ -248,7 +248,7 @@ do_paste(const char *text)
     }
     *output++ = '\0';
 
-    display_insert_at_cursor(&v->display, clean_text);
+    display_insert(&v->display, cursor_start, cursor_end, clean_text);
 }
 
 
@@ -376,7 +376,7 @@ do_sto(int index)
 
 
 void
-do_expression(int function, int arg, int cursor)
+do_expression(int function, int arg, int cursor_start, int cursor_end)
 {
     char buf[MAXLINE];
     MPNumber *ans;
@@ -398,7 +398,7 @@ do_expression(int function, int arg, int cursor)
     
     display_push(&v->display);
 
-    display_set_cursor(&v->display, cursor);
+    display_set_cursor(&v->display, cursor_start);
     ans = display_get_answer(&v->display);
 
     ui_set_statusbar("", "");
@@ -420,7 +420,7 @@ do_expression(int function, int arg, int cursor)
 
         case FN_SHIFT:
             do_shift(arg);
-            return;
+            break;
 
         case FN_SET_ACCURACY:
             do_accuracy(arg);
@@ -448,7 +448,7 @@ do_expression(int function, int arg, int cursor)
             return;
         
         case FN_PASTE:
-            do_paste((const char *)arg); // FIXME: Probably not 64 bit safe
+            do_paste(cursor_start, cursor_end, (const char *)arg); // FIXME: Probably not 64 bit safe
             return;
         
         case FN_INSERT_CHARACTER:
@@ -465,19 +465,19 @@ do_expression(int function, int arg, int cursor)
 
         case FN_RECALL:
             SNPRINTF(buf, MAXLINE, "R%d", arg);
-            display_insert_at_cursor(&v->display, buf);
+            display_insert(&v->display, cursor_start, cursor_end, buf);
             break;
 
         case FN_CONSTANT:
-            display_insert_number_at_cursor(&v->display, constant_get_value(arg));
+            display_insert_number(&v->display, cursor_start, cursor_end, constant_get_value(arg));
             break;
 
         case FN_BACKSPACE:
-            display_backspace(&v->display);
+            display_backspace(&v->display, cursor_start, cursor_end);
             break;
         
         case FN_DELETE:
-            display_delete(&v->display);
+            display_delete(&v->display, cursor_start, cursor_end);
             break;
 
         case FN_CHANGE_SIGN:
@@ -509,7 +509,6 @@ do_expression(int function, int arg, int cursor)
              * the cursor must be taken from the first undo */
             if (display_is_result(&v->display)) {
                 display_pop(&v->display);
-                cursor = display_get_cursor(&v->display);
                 if (display_is_undo_step(&v->display)) {
                     display_pop(&v->display);
                 }
@@ -560,6 +559,18 @@ do_expression(int function, int arg, int cursor)
                        message = _("Overflow. Try a bigger word size");
                        break;
 
+                    case -PARSER_ERR_UNKNOWN_VARIABLE:
+                        /* Translators; Error displayd to user when they
+                         * an unknown variable is entered */
+                       message = _("Unknown variable");
+                       break;
+
+                    case -PARSER_ERR_UNKNOWN_FUNCTION:
+                        /* Translators; Error displayd to user when they
+                         * an unknown function is entered */
+                       message = _("Unknown function");
+                       break;
+
                     case -MPMATH_ERR:
                         message = v->math_error_text;
                         break;
@@ -583,9 +594,9 @@ do_expression(int function, int arg, int cursor)
             } else {
                 if (functions[function].flags & FUNC) {
                     SNPRINTF(buf, MAXLINE, "%s(", functions[function].symname);
-                    display_insert_at_cursor(&v->display, buf);
+                    display_insert(&v->display, cursor_start, cursor_end, buf);
                 } else {
-                    display_insert_at_cursor(&v->display, functions[function].symname);
+                    display_insert(&v->display, cursor_start, cursor_end, functions[function].symname);
                 }
             }
             break;
